import os
import site
import socket
import subprocess

from fhempy.lib.generic import FhemModule

from .. import fhem


class esphome(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.proc = None
        self._set_list = {"start": {}, "stop": {}, "restart": {}}
        self.set_set_config(self._set_list)
        self._attr_list = {
            "disable": {"default": "0", "options": "0,1"},
            "port_dashboard": {"default": "6052", "help": "Default port ist 6052"},
        }
        self.set_attr_config(self._attr_list)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        if self._attr_disable == "1":
            return

        await self.start_process()

        if await fhem.init_done(hash) == 1:
            # create weblinks on first define
            self.create_async_task(self.create_weblink())

    async def start_process(self):
        my_env = os.environ
        my_env["PATH"] = site.getuserbase() + "/bin:" + my_env["PATH"]
        self._esphomeargs = [
            site.getuserbase() + "/bin/esphome",
            "dashboard",
            "esphome_config/",
            "--port",
            self._attr_port_dashboard,
        ]

        try:
            self.proc = subprocess.Popen(self._esphomeargs, env=my_env)
        except Exception:
            self.logger.exception("Failed to execute esphome")
            try:
                self._esphomeargs = [
                    "esphome",
                    "dashboard",
                    "esphome_config/",
                    "--port",
                    self._attr_port_dashboard,
                ]
                self.proc = subprocess.Popen(self._esphomeargs)
            except Exception:
                self.logger.exception("Failed to execute esphome")
                return "Failed to execute esphome"

        await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)

    async def stop_process(self):
        if self.proc:
            self.proc.kill()
            self.proc = None
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def create_weblink(self):
        if await fhem.checkIfDeviceExists(
            self.hash, "TYPE", "weblink", "NAME", "esphome_dashboard"
        ):
            return

        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        await fhem.CommandDefine(
            self.hash, "esphome_dashboard weblink iframe http://" + local_ip + ":6052/"
        )
        await fhem.CommandAttr(
            self.hash,
            (
                "esphome_dashboard htmlattr width='900' height='700' "
                "frameborder='0' marginheight='0' marginwidth='0'"
            ),
        )
        await fhem.CommandAttr(self.hash, "esphome_dashboard room ESPHome")
        await fhem.CommandAttr(self.hash, "esphome_dashboard sortby 2")

        await fhem.CommandDefine(
            self.hash,
            """
esphome_installer weblink htmlCode <style>a[target='_blank']::after {
content: '';
width: 1em;
height: 1em;
margin: 0 0.05em 0 0.1em;
background: url(data:image/svg+xml;base64,
PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb
3g9IjAgMCAxNiAxNiIgd2lkdGg9IjE2IiBoZWlnaHQ9IjE2Ij48cGF0aCBkPS
JNOSAyTDkgMyAxMi4zIDMgNiA5LjMgNi43IDEwIDEzIDMuNyAxMyA3IDE0IDc
gMTQgMlpNNCA0QzIuOSA0IDIgNC45IDIgNkwyIDEyQzIgMTMuMSAyLjkgMTQg
NCAxNEwxMCAxNEMxMS4xIDE0IDEyIDEzLjEgMTIgMTJMMTIgNyAxMSA4IDExI
DEyQzExIDEyLjYgMTAuNiAxMyAxMCAxM0w0IDEzQzMuNCAxMyAzIDEyLjYgMy
AxMkwzIDZDMyA1LjQgMy40IDUgNCA1TDggNSA5IDRaIi8+PC9zdmc+) no-repeat;
background-size: contain;
display: inline-block;
vertical-align: text-bottom;
}</style>
<a href='https://esphome.github.io/esp-web-tools/'
 style='font-size: 20px;' target='_blank'>
Click here to easily install ESPHome on a new devices
</a><br><br>""".replace(
                "\n", ""
            ),
        )
        await fhem.CommandAttr(self.hash, "esphome_installer room ESPHome")
        await fhem.CommandAttr(self.hash, "esphome_installer sortby 1")

    # FHEM FUNCTION
    async def Undefine(self, hash):
        await self.stop_process()
        return await super().Undefine(hash)

    async def set_attr_disable(self, hash):
        if self._attr_disable == "0":
            await self.start_process()
        else:
            await self.stop_process()

    async def set_start(self, hash, params):
        await self.stop_process()
        await self.start_process()
        return ""

    async def set_stop(self, hash, params):
        await self.stop_process()
        return ""

    async def set_restart(self, hash, params):
        return await self.set_start(hash, None)



import asyncio
import datetime
from aionefit import NefitCore

import re
import math

from .. import fhem, generic


class nefit(generic.FhemModule):

    URL_RRC_UISTATUS = "/ecus/rrc/uiStatus"
    URL_REC_GASUSAGEPOINTER = "/ecus/rrc/recordings/gasusagePointer"
    URL_REC_GASUSAGE = "/ecus/rrc/recordings/gasusage"
    URL_REC_YEARTOTAL = "/ecus/rrc/recordings/yearTotal"
    URL_OUTDOOR_TEMP = "/system/sensors/temperatures/outdoor_t1"
    URL_DAY_STARTSWITH = "/ecus/rrc/dayassunday/day"
    URL_DAY_ACTIVE = "/ecus/rrc/dayassunday/day%DAY%/active"
    URL_DAY_MODE = "/ecus/rrc/dayassunday/day%DAY%/mode"
    URL_DAY_DATE = "/ecus/rrc/dayassunday/day%DAY%/date"
    URL_SYSTEM_PRESSURE = "/system/appliance/systemPressure"

    def __init__(self, logger):
        super().__init__(logger)

        self.first_run = True
        self.module_shutdown = False

        attr_config = {
            "interval": {
                "default": 900,
                "format": "int",
                "help": "Change interval, default is 900.",
            },
            "password": {"default": ""},
            "access_key": {"default": ""},
        }
        self.set_attr_config(attr_config)

        set_config = {
            "mode": {
                "args": ["mode"],
                "argsh": ["mode"],
                "params": {"mode": {"default": "clock", "optional": False}},
                "options": "clock,manual",
            },
            "desiredTemp": {
                "args": ["temperature"],
                "params": {"temperature": {"format": "float"}},
                "options": "slider,10,0.5,30,1",
            },
            "dayAsSunday_00_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "0",
            },
            "dayAsSunday_01_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "1",
            },
            "dayAsSunday_02_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "2",
            },
            "dayAsSunday_03_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "3",
            },
            "dayAsSunday_04_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "4",
            },
            "dayAsSunday_05_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "5",
            },
            "dayAsSunday_06_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "6",
            },
            "dayAsSunday_12_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "12",
            },
            "dayAsSunday_07_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "7",
            },
            "dayAsSunday_08_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "8",
            },
            "dayAsSunday_09_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "9",
            },
            "dayAsSunday_10_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "10",
            },
            "dayAsSunday_11_active": {
                "args": ["onoff"],
                "options": "on,off",
                "function": "set_dayassunday_activate",
                "function_param": "11",
            },
            "dayAsSunday_07_date": {
                "args": ["dateval"],
                "function": "set_dayassunday_date",
                "function_param": "7",
            },
            "dayAsSunday_08_date": {
                "args": ["dateval"],
                "function": "set_dayassunday_date",
                "function_param": "8",
            },
            "dayAsSunday_09_date": {
                "args": ["dateval"],
                "function": "set_dayassunday_date",
                "function_param": "9",
            },
            "dayAsSunday_10_date": {
                "args": ["dateval"],
                "function": "set_dayassunday_date",
                "function_param": "10",
            },
            "dayAsSunday_11_date": {
                "args": ["dateval"],
                "function": "set_dayassunday_date",
                "function_param": "11",
            },
            "todayAsSunday": {"args": ["onoff"], "options": "on,off"},
            "tomorrowAsSunday": {"args": ["onoff"], "options": "on,off"},
        }
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define netfit_thermostat fhempy nefit <SERIAL_NUMBER>"

        self._serial_number = args[3]
        hash["SERIAL_NUMBER"] = self._serial_number

        if self._attr_access_key == "" or self._attr_password == "":
            await fhem.readingsBeginUpdate(hash)
            await fhem.readingsBulkUpdateIfChanged(
                hash, "state", "set access_key and password attribute"
            )
            await fhem.readingsEndUpdate(hash, 1)
        else:
            await fhem.readingsSingleUpdateIfChanged(hash, "state", "connecting", 1)
            self.create_async_task(self.nefit_connect())

    async def set_attr_password(self, hash):
        if self._attr_password != "" and self._attr_access_key != "":
            self.create_async_task(self.nefit_connect())

    async def set_attr_access_key(self, hash):
        if self._attr_password != "" and self._attr_access_key != "":
            self.create_async_task(self.nefit_connect())

    async def set_dayassunday_activate(self, hash, params):
        day = params["function_param"]
        onoff = params["onoff"]
        self._nefit_client.put_value(nefit.URL_DAY_ACTIVE.replace("%DAY%", day), onoff)
        await self.update_dayassunday(day)

    async def set_dayassunday_date(self, hash, params):
        day = params["function_param"]
        dateval = params["dateval"]
        self._nefit_client.put_value(nefit.URL_DAY_DATE.replace("%DAY%", day), dateval)
        await self.update_dayassunday(day)

    async def set_desiredTemp(self, hash, params):
        self._nefit_client.set_temperature(params["temperature"])
        self._nefit_client.get(nefit.URL_RRC_UISTATUS)

    async def set_mode(self, hash, params):
        self._nefit_client.set_usermode(params["mode"])
        self._nefit_client.get(nefit.URL_RRC_UISTATUS)

    async def set_todayAsSunday(self, hash, params):
        # call set_dayassunday_activate
        await self.set_dayassunday_activate(
            hash, {"onoff": params["onoff"], "function_param": "11"}
        )
        if params["onoff"] == "on":
            # call set_dayssunday_date
            today = datetime.date.today()
            await self.set_dayassunday_date(
                hash,
                {
                    "dateval": f"{today.month:02d}-{today.day:02d}",
                    "function_param": "11",
                },
            )

    async def set_tomorrowAsSunday(self, hash, params):
        # call set_dayassunday_activate
        await self.set_dayassunday_activate(
            hash, {"onoff": params["onoff"], "function_param": "10"}
        )
        if params["onoff"] == "on":
            # call set_dayssunday_date
            tomorrow = datetime.date.today() + datetime.timedelta(days=1)
            await self.set_dayassunday_date(
                hash,
                {
                    "dateval": f"{tomorrow.month:02d}-{tomorrow.day:02d}",
                    "function_param": "10",
                },
            )

    async def received_message(self, msg):
        try:
            if msg["id"] == nefit.URL_RRC_UISTATUS:
                await self.handle_uistatus(msg)
            elif msg["id"] == nefit.URL_REC_GASUSAGEPOINTER:
                await self.handle_gasusagepointer(msg)
            elif msg["id"] == nefit.URL_REC_GASUSAGE:
                await self.handle_gasusage(msg)
            elif msg["id"] == nefit.URL_REC_YEARTOTAL:
                await self.handle_yeartotal(msg)
            elif msg["id"] == nefit.URL_OUTDOOR_TEMP:
                await self.handle_outdoortemp(msg)
            elif msg["id"] == nefit.URL_SYSTEM_PRESSURE:
                await self.handle_systempressure(msg)
            elif msg["id"].startswith(nefit.URL_DAY_STARTSWITH):
                await self.handle_dayassunday(msg)
        except Exception:
            self.logger.exception(f"Failed to handle msg: {msg}")

    async def handle_systempressure(self, msg):
        if msg["value"] <= msg["maxValue"] and msg["value"] >= msg["minValue"]:
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "system_pressure", float(msg["value"]) / 10, 1
            )

    async def handle_dayassunday(self, msg):
        day = int(re.findall(r"\d+", msg["id"])[0])
        val_type = re.findall(r"/(\w+)$", msg["id"])[0]
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, f"dayassunday_{day:02d}_{val_type}", msg["value"], 1
        )

    async def handle_outdoortemp(self, msg):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "outdoor_temperature", msg["value"], 1
        )

    async def handle_yeartotal(self, msg):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "year_total_kwh", msg["value"], 1
        )

    async def handle_gasusage(self, msg):
        entry = msg["value"][self._gasusage_page_entry]
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
        if entry["d"] == f"{yesterday.day:02d}-{yesterday.month:02d}-{yesterday.year}":
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "yesterday_consumption_ch", entry["ch"], 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "yesterday_consumption_hw", entry["hw"], 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "yesterday_temperature", entry["T"] / 10, 1
            )

    async def handle_gasusagepointer(self, msg):
        if msg["value"] < 2:
            # no entry in history yet
            return

        self._gasusage_page = math.ceil((msg["value"] - 2) / 32)
        self._gasusage_page_entry = (msg["value"] - 2) % 32
        self._nefit_client.get(
            nefit.URL_REC_GASUSAGE + "?page=" + str(self._gasusage_page)
        )

    async def handle_uistatus(self, msg):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "ars", msg["value"]["ARS"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "boiler_heating", msg["value"]["BAI"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "boiler_block", msg["value"]["BBE"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "boiler_lock", msg["value"]["BLE"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "boiler_maintainance", msg["value"]["BMR"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "clock_program", msg["value"]["CPM"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "current_switchpoint", msg["value"]["CSP"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "current_time_date", msg["value"]["CTD"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "control", msg["value"]["CTR"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "todayAsSunday", msg["value"]["DAS"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "hot_water", msg["value"]["DHW"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "dot", msg["value"]["DOT"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "esi", msg["value"]["ESI"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temp_in_fahrenheit", msg["value"]["FAH"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "fireplace_mode", msg["value"]["FPA"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "presence_detection_status_device", msg["value"]["HED_DB"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "presence_detection_device", msg["value"]["HED_DEV"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "presence_detection", msg["value"]["HED_EN"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "holiday_mode", msg["value"]["HMD"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "inhouse_status", msg["value"]["IHS"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperature", msg["value"]["IHT"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "manual_mode_temperature", msg["value"]["MMT"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "pmr", msg["value"]["PMR"]
            )
            await fhem.readingsBulkUpdateIfChanged(self.hash, "rs", msg["value"]["RS"])
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "tomorrowAsSunday", msg["value"]["TAS"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperature_override_duration", msg["value"]["TOD"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperature_override", msg["value"]["TOR"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperature_override", msg["value"]["TOT"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "desiredTemp", msg["value"]["TSP"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "mode", msg["value"]["UMD"]
            )
        except Exception:
            self.logger.exception("Failed to handle uiStatus")

        await fhem.readingsEndUpdate(self.hash, 1)

    async def nefit_connect(self):
        if self.first_run:
            self._nefit_client = NefitCore(
                self._serial_number,
                self._attr_access_key,
                self._attr_password,
                message_callback=self.received_message,
            )

            self._nefit_client.failed_auth_handler = self.failed_auth_handler
            self._nefit_client.no_content_callback = self.no_content_callback
            self._nefit_client.session_end_callback = self.session_end_callback

        await self._nefit_client.connect()
        await self._nefit_client.xmppclient.connected_event.wait()

        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected", 1)

        if self.first_run:
            self.create_async_task(self.update_loop())
            self.first_run = False

    async def failed_auth_handler(self, event):
        pass

    async def no_content_callback(self, data):
        pass

    async def session_end_callback(self):
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "disconnected", 1)
        if not self.module_shutdown:
            await asyncio.sleep(10)
            await self.nefit_connect()

    async def update_loop(self):
        while True:
            try:
                self._nefit_client.get(nefit.URL_RRC_UISTATUS)
                self._nefit_client.get(nefit.URL_REC_YEARTOTAL)
                self._nefit_client.get(nefit.URL_OUTDOOR_TEMP)
                self._nefit_client.get(nefit.URL_SYSTEM_PRESSURE)
                await self.update_dayassunday()

                await self.update_gasusage()
            except Exception:
                self.logger.exception("Failed to update readings")
            try:
                await asyncio.sleep(self._attr_interval)
            except asyncio.CancelledError:
                pass

    async def update_dayassunday(self, day=None):
        if day is None:
            for day in range(13):
                self._nefit_client.get(nefit.URL_DAY_ACTIVE.replace("%DAY%", str(day)))
                self._nefit_client.get(nefit.URL_DAY_DATE.replace("%DAY%", str(day)))
                self._nefit_client.get(nefit.URL_DAY_MODE.replace("%DAY%", str(day)))
        else:
            self._nefit_client.get(nefit.URL_DAY_ACTIVE.replace("%DAY%", str(day)))
            self._nefit_client.get(nefit.URL_DAY_DATE.replace("%DAY%", str(day)))
            self._nefit_client.get(nefit.URL_DAY_MODE.replace("%DAY%", str(day)))

    async def update_gasusage(self):
        self._nefit_client.get(nefit.URL_REC_GASUSAGEPOINTER)

    async def Undefine(self, hash):
        await super().Undefine(hash)
        self.module_shutdown = True



import asyncio
import functools
import os
import shutil
import signal
import subprocess

import netifaces
from fhempy.lib.generic import FhemModule
from git import Repo

from .. import fhem, utils


class zigbee2mqtt(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.proc = None
        self._set_list = {"start": {}, "stop": {}, "restart": {}, "update": {}}
        self.set_set_config(self._set_list)
        attr_conf = {"disable": {"options": "0,1", "default": "0", "format": "int"}}
        self.set_attr_config(attr_conf)

        self.check_process_task = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        if self._attr_disable == 1:
            return

        self.create_async_task(self.run_z2m())

    async def run_z2m(self):
        res = await self.check_node_installation()
        if res is False:
            return

        res = await self.check_npm_installation()
        if res is False:
            return

        res = await self.install_z2m()
        if res is False:
            return

        await self.start_process()

    async def update_z2m(self):
        try:
            await fhem.readingsSingleUpdate(
                self.hash, "update", "started, may take a few minutes", 1
            )
            current_directory = os.getcwd()
            z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")

            await self.stop_process()
            try:
                await utils.run_blocking(
                    functools.partial(
                        shutil.copytree,
                        z2m_directory + "/data",
                        z2m_directory + "/data-backup",
                    )
                )
            except FileExistsError:
                shutil.rmtree(z2m_directory + "/data-backup")
                await utils.run_blocking(
                    functools.partial(
                        shutil.copytree,
                        z2m_directory + "/data",
                        z2m_directory + "/data-backup",
                    )
                )

            rep = Repo(z2m_directory)
            await utils.run_blocking(
                functools.partial(rep.git.checkout, "HEAD", "--", "npm-shrinkwrap.json")
            )
            o = rep.remotes.origin
            await utils.run_blocking(functools.partial(o.pull))

            await utils.run_blocking(
                functools.partial(subprocess.call, ["npm", "ci"], cwd=z2m_directory)
            )

            source_folder = z2m_directory + "/data-backup"
            destination_folder = z2m_directory + "/data"
            for file in os.listdir(source_folder):
                file_path = os.path.join(source_folder, file)
                if os.path.isfile(file_path):
                    await utils.run_blocking(
                        functools.partial(shutil.copy, file_path, destination_folder)
                    )

            await fhem.readingsSingleUpdate(self.hash, "update", "successful", 1)
        except Exception:
            self.logger.exception("Failed to update")
            await fhem.readingsSingleUpdate(
                self.hash, "update", "failed to update, check log", 1
            )

    async def install_z2m(self):
        GIT_URL = "https://github.com/Koenkk/zigbee2mqtt.git"

        current_directory = os.getcwd()
        install_directory = os.path.join(current_directory, r".fhempy")
        if not os.path.exists(install_directory):
            os.makedirs(install_directory)
        z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")
        clone_needed = True
        if not os.path.exists(z2m_directory):
            os.makedirs(z2m_directory)
        else:
            try:
                Repo(z2m_directory)
                clone_needed = False
            except Exception:
                pass

        try:
            if clone_needed:
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "installation",
                    "downloading zigbee2mqtt",
                    1,
                )
                await utils.run_blocking(
                    functools.partial(Repo.clone_from, GIT_URL, z2m_directory)
                )

            if (
                await fhem.ReadingsVal(self.hash["NAME"], "installation", "nok")
                != "successful"
            ):
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "installation",
                    "installing dependencies, might take some minutes",
                    1,
                )
                await utils.run_blocking(
                    functools.partial(subprocess.call, ["npm", "ci"], cwd=z2m_directory)
                )

                z2m_conf = (
                    "homeassistant: false\n"
                    "permit_join: false\n"
                    "mqtt:\n"
                    "  base_topic: zigbee2mqtt\n"
                    "  server: 'mqtt://localhost'\n"
                    "  client_id: 'zigbee_pi'\n"
                    "serial:\n"
                    "  port: /dev/ttyUSB0\n"
                    "frontend:\n"
                    "  port: 8080\n"
                )
                with open(z2m_directory + "/data/configuration.yaml", "w") as file:
                    file.writelines(z2m_conf)

                await self.create_weblink()
                await fhem.readingsSingleUpdate(
                    self.hash, "installation", "successful", 1
                )
        except Exception:
            self.logger.exception("Failed to clone repo")
            await fhem.readingsSingleUpdate(
                self.hash,
                "state",
                "failed to install, check fhempy log",
                1,
            )
            return False

        await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)

        return True

    async def check_node_installation(self):
        try:
            ver = await utils.run_blocking(
                functools.partial(subprocess.check_output, ["node", "--version"])
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "node", ver.decode("ascii").rstrip(), 1
            )
            return True
        except Exception:
            await fhem.readingsSingleUpdate(
                self.hash, "node", "nodejs installation missing", 1
            )
            return False

    async def check_npm_installation(self):
        try:
            ver = await utils.run_blocking(
                functools.partial(subprocess.check_output, ["npm", "--version"])
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "npm", ver.decode("ascii").rstrip(), 1
            )
            return True
        except Exception:
            await fhem.readingsSingleUpdate(
                self.hash, "npm", "npm installation missing", 1
            )
            return False

    async def start_process(self):
        current_directory = os.getcwd()
        z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")

        try:
            self.proc = subprocess.Popen(["node", "./index.js"], cwd=z2m_directory)
            await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)
            if self.check_process_task is None:
                self.check_process_task = self.create_async_task(self.check_process())
        except Exception:
            self.logger.exception("Failed to start zigbee2mqtt with npm start")

    async def check_process(self):
        while True:
            try:
                if self.proc is not None:
                    poll = self.proc.poll()
                    if poll is None:
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "state", "running", 1
                        )
                    elif poll != 0:
                        await fhem.readingsSingleUpdate(
                            self.hash, "state", "error, check log file", 1
                        )
                        await asyncio.sleep(10)
                        await self.start_process()
                    elif poll == 0:
                        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)
                        await asyncio.sleep(10)
                        await self.start_process()
                await asyncio.sleep(10)
            except asyncio.CancelledError:
                return

    async def stop_process(self):
        if self.proc:
            self.proc.send_signal(signal.SIGINT)
            self.proc = None
        if self.check_process_task:
            self.cancel_async_task(self.check_process_task)
            self.check_process_task = None
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def create_weblink(self):
        ip_list = [
            netifaces.ifaddresses(iface)[netifaces.AF_INET][0]["addr"]
            for iface in netifaces.interfaces()
            if netifaces.AF_INET in netifaces.ifaddresses(iface) and iface != "lo"
        ]
        await fhem.CommandDefine(
            self.hash, "z2m_frontend weblink iframe http://" + ip_list[0] + ":8080/"
        )
        await fhem.CommandAttr(
            self.hash,
            (
                "z2m_frontend htmlattr width='100%' height='90%' "
                "frameborder='0' marginheight='0' marginwidth='0'"
            ),
        )
        await fhem.CommandAttr(self.hash, "z2m_frontend room Zigbee2MQTT")

    # FHEM FUNCTION
    async def Undefine(self, hash):
        await self.stop_process()
        return await super().Undefine(hash)

    async def set_attr_disable(self, hash):
        if self._attr_disable == 0:
            await self.start_process()
        else:
            await self.stop_process()

    async def set_start(self, hash, params):
        await self.stop_process()
        await self.start_process()
        return ""

    async def set_stop(self, hash, params):
        await self.stop_process()
        return ""

    async def set_restart(self, hash, params):
        return await self.set_start(hash, None)

    async def set_update(self, hash, params):
        self.create_async_task(self.update_z2m())



import json
import logging
import os
import sys
from unittest.mock import MagicMock

import pytest
from fhempy.lib import fhem_pythonbinding
from fhempy.lib.pkg_installer import is_installed

from ..utils import mock_fhem

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_update(mocker):

    # monkey patch MagicMock
    async def async_magic():
        pass

    MagicMock.__await__ = lambda x: async_magic().__await__()

    mock_fhem.mock_module(mocker)
    websocket = MagicMock()
    fhempy = fhem_pythonbinding.PyBinding(websocket)

    update_msg = {
        "id": "123",
        "msgtype": "update",
        "NAME": "testdevice",
        "PYTHONTYPE": "testname",
        "FHEMPYTYPE": "testname",
        "args": [],
        "argsh": {},
        "defargs": [],
        "defargsh": {},
    }
    update_text = json.dumps(update_msg)
    await fhempy.onMessage(update_text)

    assert fhem_pythonbinding.exit_code == 1
    assert fhem_pythonbinding.stop_event.is_set() is True
    assert is_installed("fhempy")
